/*	Justin Fiedler
	MyGlWindow.cpp

	Handles most of the Graphics in a Gl_Window that interacts with the UI of main.cpp.
	Uses interface from Cubeit.cpp and Cubeit3.cpp however vastly reordered to make it
	more comprehensiable.  This file also keeps track of all the control points, track
	points and axis data for constructing the track.
*/

#include "MyGlWindow.h"

//A vector of control points for curves
std::vector<point> cPoints;
//A vector of track points for the coaster
std::vector<point> tPoints;
//A vector of vectors to handle proper car rotation
std::vector<point> rPoints;

int selectedPoint = -1;
static int last_push;

//MyGlWindow constructor
MyGlWindow::MyGlWindow(int x, int y, int w, int h) : Fl_Gl_Window(x,y,w,h)
{	//set the correct GL mode
	mode( FL_RGB | FL_ALPHA | FL_DOUBLE | FL_STENCIL );

	//start the camera somewhere
	eyeZ = 20;
	fieldOfView = 45;
	ball.reset(.3f,0,0);

	//mode variables set mode variables
	animate = false;
	mainView = true;
	arcMode = false;
	drag = false;
}

//draw the world
void MyGlWindow::draw()
//==========================================================================
{
  //set background color
  glClearColor(0,0,.3f,0);
  // clear out stencil buffer for shadows
  glClearStencil(0);
  // clear buffers (color, depth, stencil)
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  
  
  // set up the drawing transforms
  setProjection();

  //draw the ground plane
  setupFloor();
  drawFloor();

  //draw shadows
  setupShadows();
  //draw objects without color for shadows
  drawPoints(0);
  drawArcsp(0);
  drawCars(0);
  drawTrees(0);
  unsetupShadows();

  //draw the objects
  drawPoints();
  if(drag) drawArcsc();  //caluclate points but dont save them yet (this increases performance
						 //by decreasing vector adds/deletes)
  drawArcsp();
  drawCars();
  drawTrees();
}

//set up the view projection matrix for selection
void MyGlWindow::setProjection(int clearProjection)
//==========================================================================
{
  glMatrixMode(GL_PROJECTION);
  glViewport(0,0,w(),h());
  if (clearProjection)
	glLoadIdentity();
  // compute the aspect ratio so we don't distort things
  double aspect = ((double) w()) / ((double) h());
  gluPerspective(fieldOfView, aspect, .1, 100);

  // put the camera where we want it to be
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  // use the transformation in the ArcBall
  if(mainView) {
	glTranslated(0, 0, -eyeZ);
	ball.multMatrix();
  } else { //Train View
		float y_mod = 1.2*trainCars[0].scale;
		float z_mod = 0;//-.75*t->scale;
		point p, np;
		p.setp(tPoints[trainCars[0].curPoint]);
		trainCars[0].getRotation(arcMode);
		switch(arcMode){
			case ARC_LENGTH:
				np.setp(tPoints[(trainCars[0].curPoint + (int)(trainCars[0].length/2))%tPoints.size()]);
				break;
			default:
				np.setp(tPoints[(trainCars[0].curPoint + 1)%tPoints.size()]);
		};
		gluLookAt(	p.x,		p.y + y_mod,		p.z + z_mod,
					np.x,		np.y + y_mod,		np.z + z_mod,
					trainCars[0].y_vect.x, trainCars[0].y_vect.y, trainCars[0].y_vect.z);
  }
}


//**************************************************************************
//
// Handle whatever events come our way
//
// The mouse behavior is:
//    right mouse = control view. Basically, it orbits around the center
//    (if we're looking in 3D)
//    we'll remember where the mouse goes down for later
//
// an issue: the fl event for dragging is the last button pressed 
// so using the shift key can mess you up - we'll remember what the
// last push event was, so we can do the right thing
//==========================================================================
int MyGlWindow::handle(int e)
//==========================================================================
{
  switch(e) {
  case FL_SHOW:		// you must handle this, or not be seen!
	  show();
	  return 1;
  case FL_PUSH:
	  switch(Fl::event_button()) {
		case 1:  // left mouse button down is "select" - do picking
			doPick();
			last_push = 1;
		  break;
		case 3: // right mouse button down
			if(mainView) {
				float x,y;
				getMouseNDC(x,y);
				ball.down(x,y);
				last_push = 3;
			}
			break;
	  };
	  damage(1);
	  return 1;
  case FL_RELEASE:
	  //if the user is done moving the control point generate the new track points
	  if(drag) {
		genAllArcPoints(arcMode);
		drag = !drag;
	  }
	  damage(1);
	  return 1;
  case FL_DRAG: // if the user drags the mouse
	  switch(last_push /*Fl::event_button()*/) {
	  case 1: // drag with the left mouse
		if (selectedPoint > -1) {
			drag = true;
			double r1x, r1y, r1z, r2x, r2y, r2z;
			getMouseLine(r1x,r1y,r1z, r2x,r2y,r2z);

			double rx, ry, rz;
			mousePoleGo(r1x,r1y,r1z, r2x,r2y,r2z, 
						static_cast<double>(cPoints[selectedPoint].x), 
						static_cast<double>(cPoints[selectedPoint].y),
						static_cast<double>(cPoints[selectedPoint].z),
						rx, ry, rz,
						(Fl::event_state() & FL_CTRL) != 0);
			cPoints[selectedPoint].x = (float) rx;
			cPoints[selectedPoint].y = (float) ry;
			cPoints[selectedPoint].z = (float) rz;
			drawArcsc();	//only calculate the arc but dont save it yet
			damage(1);
		};
		break;
	  case 3: // right mouse button
		float x,y;
		getMouseNDC(x,y);
		ball.computeNow(x,y);
		damage(1);
	  }
	  return 1;
  
  case FL_KEYBOARD:
	  return 0;
  case FL_MOUSEWHEEL:{
	  //zoom in and out on mouse scroll wheel
	  int scroll = Fl::event_dy();
	  eyeZ += scroll;
	  damage(1);
	  break;
	}
  default:
	return 0;
  }
}


// this has to be a method since we need access to make_current
void MyGlWindow::doPick()
//==========================================================================
{
	int mx = Fl::event_x(); // where is the mouse?
	int my = Fl::event_y();
	make_current();		// so we can do drawing outside of
						// draw method
	// we need the size of the OpenGL window, in its own terms
	int viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);
	// set up the pick matrix on the stack - remember, FlTk is
	// upside down!
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	gluPickMatrix((double)mx, (double)(viewport[3]-my), 5, 5, viewport);
	// now, setup the drawing transforms
	setProjection(0);
	// now draw the objects - but really only see what we hit
	GLuint buf[100];
	glSelectBuffer(100,buf);
	glRenderMode(GL_SELECT);
	glInitNames();
	glPushName(0);
	drawPoints(0,1);
	// go back to drawing mode, and see how picking did
	int hits = glRenderMode(GL_RENDER);
	if (hits) {
		// warning; this just grabs the first object hit - if there
		// are multiple objects, you really want to pick the closest
		// one - see the OpenGL manual 
		// remember: we load names that are one more than the index
		selectedPoint = buf[3]-1;
	} else // nothing hit, nothing selected
		selectedPoint = -1;
}


// get mouse position in NDC
void MyGlWindow::getMouseNDC(float& x, float& y)
{	  float mx = (float) Fl::event_x();	// remeber where the mouse went down
	  float my = (float) Fl::event_y();

	  //assume that the viewport is the same as the window size
	  float wd = (float) w();
	  float hd = (float) h();

	  //FlTk has Y going the wrong way!
	  my = hd-my;
	
	  x = (mx / wd) * 2.0f - 1.f;
	  y = (my / hd) * 2.0f - 1.f;
}

//draw the points
void MyGlWindow::drawPoints(int doColors, int doNames)
{	for(int ct=0; ct< (int) cPoints.size(); ct++) {
	if (doColors) {
		if (selectedPoint == ct){
			glColor3f(1,1,0);
		} else {
			glColor3f(1,0,0);
		}
	}
	if (doNames) glLoadName(ct + 1);
	drawCube(cPoints[ct].x, cPoints[ct].y, cPoints[ct].z, .25);
  }
}

//draw the train cars
void MyGlWindow::drawCars(int doColors, int doNames)
{	Passenger pass;
	for(int i = 0; i < trainCars.size(); i++){
		trainCars[i].draw(doColors, arcMode);
	}
}

//draw the points
void MyGlWindow::drawTrees(int doColors, int doNames)
{	for(int ct=0; ct< (int) trees.size(); ct++) {
		if (doNames) glLoadName(ct + 1);
		trees[ct].draw(doColors);
	}
}

// Callback for X dimension
void xRollerCB(Fl_Widget* w, void* p){
	if (selectedPoint > -1) {
		cPoints[selectedPoint].x += (float) static_cast<Fl_Roller*>(w)->value();
		static_cast<Fl_Roller*>(w)->value(0);
		if (p) static_cast<Fl_Widget*>(p)->damage(1);
	}
}

// Callback for Y dimension
void yRollerCB(Fl_Widget* w, void* p){
	if (selectedPoint > -1) {
		cPoints[selectedPoint].y += (float) static_cast<Fl_Roller*>(w)->value();
		static_cast<Fl_Roller*>(w)->value(0);
		if (p) static_cast<Fl_Widget*>(p)->damage(1);
	}
}

// Callback for Z dimension
void zRollerCB(Fl_Widget* w, void* p){
	if (selectedPoint > -1) {
		cPoints[selectedPoint].z += (float) static_cast<Fl_Roller*>(w)->value();
		static_cast<Fl_Roller*>(w)->value(0);
		if (p) static_cast<Fl_Widget*>(p)->damage(1);
	}
}

/* Need a callback for each dimension
 * the reseting each time doesn't really work, but its a cheesy
 * way to avoid having to reset the rollers when a new object is selected
 */
void makeRollers(int x, int y, int h, Fl_Widget* w){
	Fl_Roller* rx = new Fl_Roller(x,y,25,h,"X");
	rx->minimum(-1);
	rx->maximum(1);
	rx->callback(xRollerCB,w);
	Fl_Roller* ry = new Fl_Roller(x+40,y,25,h,"Y");
	ry->minimum(-1);
	ry->maximum(1);
	ry->callback(yRollerCB,w);
	Fl_Roller* rz = new Fl_Roller(x+80,y,25,h,"Z");
	rz->minimum(-1);
	rz->maximum(1);
	rz->callback(zRollerCB,w);
}